@@ -26,6 +26,9 @@
 #include FT_STROKER_H
 #include FT_GLYPH_H
 #include FT_SYNTHESIS_H
+#ifdef CONFIG_HARFBUZZ
+#include "hb.h"
+#endif
 
 // XXX: fix the inclusion mess so we can avoid doing this here
 typedef struct ass_shaper ASS_Shaper;
@@ -38,9 +41,12 @@ typedef struct ass_shaper ASS_Shaper;
 #include "ass_fontconfig.h"
 #include "ass_library.h"
 #include "ass_drawing.h"
+#include "ass_bitmap.h"
+#include "ass_rasterizer.h"
 
-#define GLYPH_CACHE_MAX 1000
-#define BITMAP_CACHE_MAX_SIZE 30 * 1048576
+#define GLYPH_CACHE_MAX 10000
+#define BITMAP_CACHE_MAX_SIZE 500 * 1048576
+#define COMPOSITE_CACHE_MAX_SIZE 500 * 1048576
 
 #define PARSED_FADE (1<<0)
 #define PARSED_A    (1<<1)
@@ -65,6 +71,8 @@ typedef struct free_list {
 typedef struct {
     int frame_width;
     int frame_height;
+    int storage_width;          // video width before any rescaling
+    int storage_height;         // video height before any rescaling
     double font_size_coeff;     // font size multiplier
     double line_spacing;        // additional line spacing (in frame pixels)
     double line_position;       // vertical position for subtitles, 0-100 (0 = no change)
@@ -74,10 +82,10 @@ typedef struct {
     int right_margin;
     int use_margins;            // 0 - place all subtitles inside original frame
     // 1 - use margins for placing toptitles and subtitles
-    double aspect;              // frame aspect ratio, d_width / d_height.
-    double storage_aspect;      // pixel ratio of the source image
+    double par;                 // user defined pixel aspect ratio (0 = unset)
     ASS_Hinting hinting;
     ASS_ShapingLevel shaper;
+    int selective_style_overrides;
 
     char *default_font;
     char *default_family;
@@ -99,6 +107,50 @@ typedef enum {
     EF_KARAOKE_KO
 } Effect;
 
+// describes a combined bitmap
+typedef struct {
+    Bitmap *bm;                 // glyphs bitmap
+    unsigned w;
+    unsigned h;
+    Bitmap *bm_o;               // outline bitmap
+    unsigned o_w;
+    unsigned o_h;
+    Bitmap *bm_s;               // shadow bitmap
+    FT_Vector pos;
+    uint32_t c[4];              // colors
+    FT_Vector advance;          // 26.6
+    Effect effect_type;
+    int effect_timing;          // time duration of current karaoke word
+    // after process_karaoke_effects: distance in pixels from the glyph origin.
+    // part of the glyph to the left of it is displayed in a different color.
+    int be;                     // blur edges
+    double blur;                // gaussian blur
+    double shadow_x;
+    double shadow_y;
+    double frx, fry, frz;       // rotation
+    double fax, fay;            // text shearing
+    double scale_x, scale_y;
+    int border_style;
+    int has_border;
+    double border_x, border_y;
+    double hspacing;
+    unsigned italic;
+    unsigned bold;
+    int flags;
+    int shift_x, shift_y;
+
+    unsigned has_outline;
+    unsigned is_drawing;
+
+    int max_str_length;
+    int str_length;
+    unsigned chars;
+    char *str;
+    int cached;
+    FT_Vector pos_orig;
+    int first_pos_x;
+} CombinedBitmapInfo;
+
 // describes a glyph
 // GlyphInfo and TextInfo are used for text centering and word-wrapping operations
 typedef struct glyph_info {
@@ -107,6 +159,11 @@ typedef struct glyph_info {
     ASS_Font *font;
     int face_index;
     int glyph_index;
+#ifdef CONFIG_HARFBUZZ
+    hb_script_t script;
+#else
+    int script;
+#endif
     double font_size;
     ASS_Drawing *drawing;
     FT_Outline *outline;
@@ -121,6 +178,7 @@ typedef struct glyph_info {
     uint32_t c[4];              // colors
     FT_Vector advance;          // 26.6
     FT_Vector cluster_advance;
+    char effect;                // the first (leading) glyph of some effect ?
     Effect effect_type;
     int effect_timing;          // time duration of current karaoke word
     // after process_karaoke_effects: distance in pixels from the glyph origin.
@@ -134,6 +192,7 @@ typedef struct glyph_info {
     double frx, fry, frz;       // rotation
     double fax, fay;            // text shearing
     double scale_x, scale_y;
+    double orig_scale_x, orig_scale_y; // scale_x,y before fix_glyph_scaling
     int border_style;
     double border_x, border_y;
     double hspacing;
@@ -160,9 +219,12 @@ typedef struct {
     int length;
     LineInfo *lines;
     int n_lines;
+    CombinedBitmapInfo *combined_bitmaps;
+    unsigned n_bitmaps;
     double height;
     int max_glyphs;
     int max_lines;
+    unsigned max_bitmaps;
 } TextInfo;
 
 // Renderer state.
@@ -171,6 +233,7 @@ typedef struct {
     ASS_Event *event;
     ASS_Style *style;
     int parsed_tags;
+    int has_clips;              // clips that conflict with cache change detection
 
     ASS_Font *font;
     double font_size;
@@ -204,7 +267,8 @@ typedef struct {
     double blur;                // gaussian blur
     double shadow_x;
     double shadow_y;
-    int drawing_mode;           // not implemented; when != 0 text is discarded, except for style override tags
+    int drawing_scale;          // currently reading: regular text if 0, drawing otherwise
+    double pbo;                 // drawing baseline offset
     ASS_Drawing *drawing;       // current drawing
     ASS_Drawing *clip_drawing;  // clip vector
     int clip_drawing_mode;      // 0 = regular clip, 1 = inverse clip
@@ -231,6 +295,12 @@ typedef struct {
     int treat_family_as_pattern;
     int wrap_style;
     int font_encoding;
+
+    // combination of ASS_OVERRIDE_BIT_* flags that apply right now
+    unsigned overrides;
+
+    // used to store RenderContext.style when doing selective style overrides
+    ASS_Style override_style_temp_storage;
 } RenderContext;
 
 typedef struct {
@@ -240,8 +310,20 @@ typedef struct {
     Cache *composite_cache;
     size_t glyph_max;
     size_t bitmap_max_size;
+    size_t composite_max_size;
 } CacheStore;
 
+typedef void (*BitmapBlendFunc)(uint8_t *dst, intptr_t dst_stride,
+                                uint8_t *src, intptr_t src_stride,
+                                intptr_t height, intptr_t width);
+typedef void (*BitmapMulFunc)(uint8_t *dst, intptr_t dst_stride,
+                              uint8_t *src1, intptr_t src1_stride,
+                              uint8_t *src2, intptr_t src2_stride,
+                              intptr_t width, intptr_t height);
+typedef void (*BEBlurFunc)(uint8_t *buf, intptr_t w,
+                           intptr_t h, intptr_t stride,
+                           uint16_t *tmp);
+
 struct ass_renderer {
     ASS_Library *library;
     FT_Library ftlibrary;
@@ -269,13 +351,24 @@ struct ass_renderer {
     double font_scale;
     double font_scale_x;        // x scale applied to all glyphs to preserve text aspect ratio
     double border_scale;
+    double blur_scale;
 
     RenderContext state;
     TextInfo text_info;
     CacheStore cache;
 
+#if CONFIG_RASTERIZER
+    ASS_Rasterizer rasterizer;
+#endif
+    BitmapBlendFunc add_bitmaps_func;
+    BitmapBlendFunc sub_bitmaps_func;
+    BitmapMulFunc mul_bitmaps_func;
+    BEBlurFunc be_blur_func;
+
     FreeList *free_head;
     FreeList *free_tail;
+
+    ASS_Style user_override_style;
 };
 
 typedef struct render_priv {
